import { Schema, Validator } from 'jsonschema'
// https://json-schema.org/learn/getting-started-step-by-step
const validator = new Validator()

export const validateJSON = (
    val: unknown,
    schema: Schema
): {
    isValid: boolean
    errorMsg: string[]
} => {
    const res = validator.validate(val, schema)
    const errorMsg = res.errors.map((e) => {
        return `${e.stack}`
    })
    return {
        isValid: res.valid,
        errorMsg: errorMsg,
    }
}
export class InvalidJSONError extends Error {
    constructor(errorDetails: {
        msg: string
        expected: Schema
        actual: unknown
    }) {
        const { msg, expected, actual } = errorDetails
        super(
            msg +
                '\nExpected:\n' +
                JSON.stringify(expected, null, 2) +
                '\nActual:\n' +
                JSON.stringify(actual, null, 2)
        )
    }
}

const PeerInfoSchemaBase = {
    type: 'object',
    properties: {
        name: { type: 'string' },
        host: { type: 'string' },
        port: { type: 'number' },
        publicKey: { type: 'string' },
        contact: { type: 'string' },
        lastSync: { type: 'number' },
        note: { type: 'string' },
    },
    required: ['name', 'host', 'port'],
}

export const SelfInfoSchema: Schema = {
    ...PeerInfoSchemaBase,
    required: ['name', 'host', 'port', 'contact'],
}

export const PeerInfoSchema: Schema = {
    ...PeerInfoSchemaBase,
}

export const PeersInfoSchema: Schema = {
    type: 'array',
    items: PeerInfoSchema,
}

export const QuestionSchema: Schema = {
    type: 'object',
    properties: {
        Q: { type: 'string' },
        A: { type: 'string' },
        data: {
            type: 'object',
            properties: {
                type: { type: 'string' },
                date: { type: 'number' },
                source: { type: 'string' },
                base64: { type: 'array', items: { type: 'string' } },
                images: { type: 'array', items: { type: 'string' } },
                hashedImages: { type: 'array', items: { type: 'string' } },
                possibleAnswers: {
                    type: 'array',
                    items: {
                        type: 'object',
                        properties: {
                            type: { type: 'string' },
                            val: { type: 'string' },
                            selectedByUser: { type: 'boolean' },
                        },
                    },
                },
            },
            required: ['type'],
        },
    },
    required: ['Q', 'A', 'data'],
}

export const SubjectSchema: Schema = {
    type: 'object',
    properties: {
        Name: { type: 'string' },
        Questions: { type: 'array', items: QuestionSchema },
    },
    required: ['Name', 'Questions'],
}

export const QuestoinDbFileSchema: Schema = {
    type: 'array',
    items: SubjectSchema,
}
